/*****************************************************************************

        InterpPack.cpp
        Copyright (c) 2003 Laurent de Soras

--- Legal stuff ---

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130) // "'operator' : logical operation on address of string constant"
	#pragma warning (1 : 4223) // "nonstandard extension used : non-lvalue array converted to pointer"
	#pragma warning (1 : 4705) // "statement has no effect"
	#pragma warning (1 : 4706) // "assignment within conditional expression"
	#pragma warning (4 : 4786) // "identifier was truncated to '255' characters in the debug information"
	#pragma warning (4 : 4800) // "forcing value to bool 'true' or 'false' (performance warning)"
	#pragma warning (4 : 4355) // "'this' : used in base member initializer list"
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"BaseVoiceState.h"
#include	"InterpPack.h"

#include	<cassert>



namespace rspl
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



InterpPack::InterpPack ()
:	_interp_1x ()
,	_interp_2x ()
{
	_interp_1x.set_impulse (_fir_1x_coef_arr);
	_interp_2x.set_impulse (_fir_2x_coef_arr);
}



void	InterpPack::interp_ovrspl (float dest_ptr [], long nbr_spl, BaseVoiceState &voice) const
{
	assert (dest_ptr != 0);
	assert (nbr_spl > 0);
	assert (&voice != 0);
	assert (voice._table_ptr != 0);

	long				cnt = 0;
	do
	{
		assert (voice._pos._part._msw < voice._table_len);

		dest_ptr [cnt] = 0.5f * _interp_2x.interpolate (
			voice._table_ptr + voice._pos._part._msw,
			voice._pos._part._lsw
		);

		voice._pos._all += voice._step._all;
		++ cnt;
	}
	while (cnt < nbr_spl);
}



void	InterpPack::interp_norm (float dest_ptr [], long nbr_spl, BaseVoiceState &voice) const
{
	assert (dest_ptr != 0);
	assert (nbr_spl > 0);
	assert (&voice != 0);
	assert (voice._table_ptr != 0);

	long				cnt = 0;
	do
	{
		assert (voice._pos._part._msw < voice._table_len);

		dest_ptr [cnt] = _interp_1x.interpolate (
			voice._table_ptr + voice._pos._part._msw,
			voice._pos._part._lsw
		);

		voice._pos._all += voice._step._all;
		++ cnt;
	}
	while (cnt < nbr_spl);
}



void	InterpPack::interp_ovrspl_ramp_add (float dest_ptr [], long nbr_spl, BaseVoiceState &voice, float vol, float vol_step) const
{
	assert (dest_ptr != 0);
	assert (nbr_spl > 0);
	assert (&voice != 0);
	assert (voice._table_ptr != 0);
	assert (vol >= 0);
	assert (vol <= 1);
	assert (vol_step >= -1);
	assert (vol_step <= 1);

	vol *= 0.5;
	vol_step *= 0.5;

	long				cnt = 0;
	do
	{
		assert (voice._pos._part._msw < voice._table_len);

		dest_ptr [cnt] += vol * _interp_2x.interpolate (
			voice._table_ptr + voice._pos._part._msw,
			voice._pos._part._lsw
		);

		voice._pos._all += voice._step._all;
		vol += vol_step;
		++ cnt;
	}
	while (cnt < nbr_spl);
}



void	InterpPack::interp_norm_ramp_add (float dest_ptr [], long nbr_spl, BaseVoiceState &voice, float vol, float vol_step) const
{
	assert (dest_ptr != 0);
	assert (nbr_spl > 0);
	assert (&voice != 0);
	assert (voice._table_ptr != 0);
	assert (vol >= 0);
	assert (vol <= 1);
	assert (vol_step >= -1);
	assert (vol_step <= 1);

	vol_step *= 2;

	long				cnt = 0;
	do
	{
		assert (voice._pos._part._msw < voice._table_len);

		dest_ptr [cnt] += vol * _interp_1x.interpolate (
			voice._table_ptr + voice._pos._part._msw,
			voice._pos._part._lsw
		);

		voice._pos._all += voice._step._all;
		vol += vol_step;
		cnt += 2;
	}
	while (cnt < nbr_spl);
}



long	InterpPack::get_len_pre ()
{
	assert (   static_cast <long> (InterpRate1x::FIR_LEN)
	        >= static_cast <long> (InterpRate2x::FIR_LEN));

	return (static_cast <long> (InterpRate1x::FIR_LEN / 2));
}



long	InterpPack::get_len_post ()
{
	assert (   static_cast <long> (InterpRate1x::FIR_LEN)
	        >= static_cast <long> (InterpRate2x::FIR_LEN));

	return (static_cast <long> (InterpRate1x::FIR_LEN / 2));
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



// Specs:
// FIR LPF
// 1 + 1535 coefficients (the first one is an additionnal 0)
// Fc = 0.5*Fs/64
// Passband frequency range: 0 - 0.85*Fc
// Passband level: +36.12 dB
// Passband ripple: 0.1 dB
// Stopband start frequency: 1.14*Fc
// Attenuation: -81/-87/-93 dB
// k = 64; b = remez (768*2-2, [0 0.85/k 1.14/k 1.99/k 2.01/k 3.99/k 4.01/k 1], [k k 0 0 0 0 0 0], [1 64 128 256]); b = [0 b];
// bp = reshape (b, 4, 768*2/4); bp = transpose (bp); dlmwrite ('pizza.txt', bp, ', ');
const double	InterpPack::_fir_1x_coef_arr [InterpRate1x::IMPULSE_LEN] =
{
	0, 0.00084549, 0.00025079, 0.00028614,
	0.00032345, 0.0003625, 0.00040326, 0.00044552,
	0.00048927, 0.00053431, 0.00058068, 0.00062815,
	0.00067673, 0.0007262, 0.00077656, 0.00082761,
	0.00087945, 0.00093197, 0.00098539, 0.0010396,
	0.0010947, 0.0011503, 0.0012067, 0.0012634,
	0.0013208, 0.0013791, 0.0014384, 0.001498,
	0.0015577, 0.0016174, 0.0016795, 0.0017398,
	0.0018011, 0.0018625, 0.0019233, 0.0019841,
	0.002044, 0.0021034, 0.0021616, 0.0022187,
	0.0022743, 0.0023281, 0.0023799, 0.0024295,
	0.0024766, 0.0025212, 0.0025628, 0.0026013,
	0.0026363, 0.0026679, 0.0026957, 0.0027199,
	0.0027399, 0.0027558, 0.0027673, 0.0027746,
	0.0027774, 0.0027755, 0.0027687, 0.0027577,
	0.0027414, 0.0027203, 0.0026943, 0.0026631,
	0.0026269, 0.0025854, 0.0025387, 0.0024866,
	0.0024291, 0.002366, 0.0022975, 0.0022233,
	0.0021434, 0.0020578, 0.0019663, 0.0018691,
	0.001766, 0.0016571, 0.0015423, 0.0014217,
	0.0012953, 0.0011633, 0.0010257, 0.00088249,
	0.00073402, 0.00058038, 0.00042169, 0.00025803,
	8.9978e-005, -8.2619e-005, -0.00025933, -0.00043984,
	-0.00062408, -0.00081155, -0.0010021, -0.0011953,
	-0.001391, -0.0015887, -0.0017882, -0.0019891,
	-0.0021911, -0.0023938, -0.0025969, -0.0028001,
	-0.0030029, -0.0032051, -0.0034062, -0.0036058,
	-0.0038038, -0.0039995, -0.0041926, -0.0043828,
	-0.0045696, -0.0047527, -0.0049315, -0.0051058,
	-0.0052751, -0.0054388, -0.0055967, -0.0057482,
	-0.005893, -0.0060306, -0.0061605, -0.0062825,
	-0.006396, -0.0065006, -0.006596, -0.0066818,
	-0.0067576, -0.0068232, -0.0068781, -0.0069221,
	-0.0069549, -0.0069763, -0.006986, -0.0069839,
	-0.0069698, -0.0069435, -0.0069049, -0.0068539,
	-0.0067904, -0.0067144, -0.0066258, -0.0065247,
	-0.0064111, -0.006285, -0.0061465, -0.0059957,
	-0.0058327, -0.0056577, -0.0054708, -0.0052722,
	-0.0050622, -0.004841, -0.0046088, -0.004366,
	-0.0041129, -0.0038499, -0.0035773, -0.0032955,
	-0.0030051, -0.0027064, -0.0023999, -0.0020863,
	-0.0017661, -0.0014398, -0.0011081, -0.00077158,
	-0.00043093, -8.6808e-005, 0.00026005, 0.00060893,
	0.00095913, 0.0013099, 0.0016604, 0.0020099,
	0.0023577, 0.0027029, 0.0030449, 0.0033826,
	0.0037155, 0.0040427, 0.0043634, 0.0046769,
	0.0049823, 0.005279, 0.0055661, 0.005843,
	0.0061088, 0.006363, 0.0066047, 0.0068334,
	0.0070483, 0.0072487, 0.0074342, 0.007604,
	0.0077575, 0.0078942, 0.0080136, 0.0081152,
	0.0081984, 0.0082628, 0.0083081, 0.0083337,
	0.0083395, 0.008325, 0.0082901, 0.0082345,
	0.008158, 0.0080605, 0.0079419, 0.0078022,
	0.0076415, 0.0074597, 0.007257, 0.0070336,
	0.0067897, 0.0065255, 0.0062415, 0.0059379,
	0.0056152, 0.0052738, 0.0049144, 0.0045374,
	0.0041434, 0.0037332, 0.0033074, 0.0028667,
	0.002412, 0.0019441, 0.0014638, 0.00097207,
	0.0004698, -4.2025e-005, -0.00056238, -0.0010903,
	-0.0016245, -0.0021642, -0.002708, -0.0032548,
	-0.0038035, -0.0043529, -0.0049017, -0.0054486,
	-0.0059925, -0.006532, -0.0070659, -0.0075928,
	-0.0081116, -0.0086207, -0.0091191, -0.0096053,
	-0.010078, -0.010536, -0.010979, -0.011404,
	-0.01181, -0.012198, -0.012564, -0.012909,
	-0.01323, -0.013528, -0.013801, -0.014048,
	-0.014268, -0.01446, -0.014624, -0.014759,
	-0.014864, -0.014938, -0.014981, -0.014993,
	-0.014972, -0.014919, -0.014833, -0.014714,
	-0.014561, -0.014376, -0.014157, -0.013904,
	-0.013619, -0.0133, -0.012948, -0.012564,
	-0.012148, -0.011701, -0.011222, -0.010714,
	-0.010175, -0.0096082, -0.0090132, -0.0083913,
	-0.0077436, -0.0070711, -0.006375, -0.0056567,
	-0.0049174, -0.0041585, -0.0033816, -0.002588,
	-0.0017795, -0.00095756, -0.00012395, 0.00071962,
	0.0015714, 0.0024295, 0.0032921, 0.0041574,
	0.0050234, 0.0058881, 0.0067497, 0.0076062,
	0.0084555, 0.0092957, 0.010125, 0.010941,
	0.011742, 0.012526, 0.01329, 0.014034,
	0.014755, 0.015451, 0.01612, 0.016761,
	0.017372, 0.01795, 0.018495, 0.019004,
	0.019476, 0.01991, 0.020304, 0.020656,
	0.020966, 0.021231, 0.021452, 0.021626,
	0.021754, 0.021833, 0.021864, 0.021845,
	0.021776, 0.021657, 0.021486, 0.021265,
	0.020993, 0.02067, 0.020295, 0.01987,
	0.019395, 0.018869, 0.018294, 0.01767,
	0.016999, 0.016281, 0.015516, 0.014707,
	0.013855, 0.012961, 0.012026, 0.011052,
	0.010042, 0.0089957, 0.0079163, 0.0068055,
	0.0056656, 0.0044987, 0.0033071, 0.0020932,
	0.00085947, -0.00039149, -0.0016571, -0.0029347,
	-0.0042215, -0.0055146, -0.0068114, -0.0081088,
	-0.0094039, -0.010694, -0.011975, -0.013246,
	-0.014502, -0.015741, -0.016959, -0.018154,
	-0.019323, -0.020463, -0.02157, -0.022642,
	-0.023677, -0.02467, -0.02562, -0.026524,
	-0.027379, -0.028183, -0.028933, -0.029627,
	-0.030263, -0.030839, -0.031352, -0.0318,
	-0.032183, -0.032498, -0.032744, -0.032919,
	-0.033023, -0.033053, -0.03301, -0.032892,
	-0.032699, -0.03243, -0.032086, -0.031665,
	-0.031168, -0.030595, -0.029947, -0.029224,
	-0.028427, -0.027556, -0.026614, -0.025601,
	-0.024518, -0.023368, -0.022152, -0.020872,
	-0.019531, -0.01813, -0.016672, -0.015159,
	-0.013596, -0.011984, -0.010326, -0.008626,
	-0.0068872, -0.0051131, -0.0033072, -0.0014732,
	0.00038505, 0.0022636, 0.0041584, 0.0060655,
	0.0079806, 0.0098996, 0.011818, 0.013732,
	0.015636, 0.017527, 0.0194, 0.02125,
	0.023074, 0.024866, 0.026622, 0.028338,
	0.03001, 0.031633, 0.033203, 0.034716,
	0.036167, 0.037553, 0.03887, 0.040114,
	0.041281, 0.042369, 0.043372, 0.044289,
	0.045116, 0.04585, 0.046488, 0.047028,
	0.047468, 0.047805, 0.048037, 0.048163,
	0.048182, 0.04809, 0.047889, 0.047577,
	0.047153, 0.046616, 0.045968, 0.045208,
	0.044336, 0.043353, 0.042261, 0.041059,
	0.039751, 0.038336, 0.036818, 0.035199,
	0.03348, 0.031665, 0.029757, 0.027758,
	0.025673, 0.023505, 0.021258, 0.018935,
	0.016543, 0.014084, 0.011563, 0.0089869,
	0.0063592, 0.0036857, 0.00097182, -0.0017767,
	-0.0045541, -0.0073545, -0.010172, -0.013,
	-0.015832, -0.018663, -0.021485, -0.024292,
	-0.027078, -0.029837, -0.03256, -0.035243,
	-0.037879, -0.04046, -0.04298, -0.045434,
	-0.047814, -0.050115, -0.05233, -0.054453,
	-0.056479, -0.058401, -0.060214, -0.061913,
	-0.063491, -0.064946, -0.06627, -0.067461,
	-0.068513, -0.069423, -0.070186, -0.070799,
	-0.071259, -0.071563, -0.071709, -0.071693,
	-0.071515, -0.071171, -0.070662, -0.069986,
	-0.069142, -0.068131, -0.066952, -0.065607,
	-0.064095, -0.062418, -0.060579, -0.058578,
	-0.056419, -0.054104, -0.051636, -0.04902,
	-0.046258, -0.043356, -0.040318, -0.037149,
	-0.033855, -0.030441, -0.026913, -0.023278,
	-0.019543, -0.015715, -0.011801, -0.0078082,
	-0.0037453, 0.00037992, 0.0045589, 0.008783,
	0.013043, 0.017331, 0.021637, 0.025951,
	0.030264, 0.034567, 0.038849, 0.043101,
	0.047313, 0.051475, 0.055577, 0.059608,
	0.063559, 0.067421, 0.071182, 0.074834,
	0.078366, 0.081769, 0.085034, 0.088151,
	0.091111, 0.093906, 0.096527, 0.098965,
	0.10121, 0.10326, 0.10511, 0.10674,
	0.10815, 0.10933, 0.11029, 0.111,
	0.11148, 0.1117, 0.11168, 0.1114,
	0.11086, 0.11006, 0.109, 0.10767,
	0.10608, 0.10422, 0.1021, 0.099704,
	0.097048, 0.094129, 0.090949, 0.087511,
	0.083819, 0.079877, 0.075689, 0.071261,
	0.066598, 0.061708, 0.056597, 0.051272,
	0.045742, 0.040016, 0.034103, 0.028013,
	0.021756, 0.015343, 0.0087852, 0.0020949,
	-0.0047158, -0.011634, -0.018646, -0.02574,
	-0.032899, -0.040111, -0.047361, -0.054632,
	-0.061911, -0.069181, -0.076426, -0.08363,
	-0.090777, -0.097851, -0.10483, -0.11171,
	-0.11846, -0.12507, -0.13153, -0.1378,
	-0.14389, -0.14977, -0.15542, -0.16083,
	-0.16599, -0.17086, -0.17545, -0.17973,
	-0.18369, -0.18731, -0.19058, -0.19349,
	-0.19601, -0.19814, -0.19987, -0.20117,
	-0.20204, -0.20247, -0.20245, -0.20196,
	-0.201, -0.19955, -0.19761, -0.19518,
	-0.19223, -0.18878, -0.18481, -0.18031,
	-0.17529, -0.16973, -0.16365, -0.15703,
	-0.14988, -0.14219, -0.13397, -0.12522,
	-0.11594, -0.10613, -0.09581, -0.08497,
	-0.07362, -0.061769, -0.049424, -0.036594,
	-0.023289, -0.0095179, 0.0047068, 0.019373,
	0.034469, 0.049979, 0.06589, 0.082187,
	0.098854, 0.11587, 0.13323, 0.1509,
	0.16887, 0.18713, 0.20564, 0.22439,
	0.24336, 0.26253, 0.28187, 0.30136,
	0.32099, 0.34072, 0.36054, 0.38042,
	0.40033, 0.42025, 0.44016, 0.46004,
	0.47985, 0.49958, 0.51919, 0.53867,
	0.55798, 0.57711, 0.59603, 0.61472,
	0.63314, 0.65129, 0.66912, 0.68663,
	0.70378, 0.72056, 0.73694, 0.7529,
	0.76842, 0.78347, 0.79805, 0.81212,
	0.82567, 0.83869, 0.85114, 0.86303,
	0.87432, 0.88501, 0.89508, 0.90452,
	0.91331, 0.92144, 0.9289, 0.93569,
	0.94178, 0.94718, 0.95187, 0.95585,
	0.95912, 0.96166, 0.96348, 0.96458,
	0.96494, 0.96458, 0.96348, 0.96166,
	0.95912, 0.95585, 0.95187, 0.94718,
	0.94178, 0.93569, 0.9289, 0.92144,
	0.91331, 0.90452, 0.89508, 0.88501,
	0.87432, 0.86303, 0.85114, 0.83869,
	0.82567, 0.81212, 0.79805, 0.78347,
	0.76842, 0.7529, 0.73694, 0.72056,
	0.70378, 0.68663, 0.66912, 0.65129,
	0.63314, 0.61472, 0.59603, 0.57711,
	0.55798, 0.53867, 0.51919, 0.49958,
	0.47985, 0.46004, 0.44016, 0.42025,
	0.40033, 0.38042, 0.36054, 0.34072,
	0.32099, 0.30136, 0.28187, 0.26253,
	0.24336, 0.22439, 0.20564, 0.18713,
	0.16887, 0.1509, 0.13323, 0.11587,
	0.098854, 0.082187, 0.06589, 0.049979,
	0.034469, 0.019373, 0.0047068, -0.0095179,
	-0.023289, -0.036594, -0.049424, -0.061769,
	-0.07362, -0.08497, -0.09581, -0.10613,
	-0.11594, -0.12522, -0.13397, -0.14219,
	-0.14988, -0.15703, -0.16365, -0.16973,
	-0.17529, -0.18031, -0.18481, -0.18878,
	-0.19223, -0.19518, -0.19761, -0.19955,
	-0.201, -0.20196, -0.20245, -0.20247,
	-0.20204, -0.20117, -0.19987, -0.19814,
	-0.19601, -0.19349, -0.19058, -0.18731,
	-0.18369, -0.17973, -0.17545, -0.17086,
	-0.16599, -0.16083, -0.15542, -0.14977,
	-0.14389, -0.1378, -0.13153, -0.12507,
	-0.11846, -0.11171, -0.10483, -0.097851,
	-0.090777, -0.08363, -0.076426, -0.069181,
	-0.061911, -0.054632, -0.047361, -0.040111,
	-0.032899, -0.02574, -0.018646, -0.011634,
	-0.0047158, 0.0020949, 0.0087852, 0.015343,
	0.021756, 0.028013, 0.034103, 0.040016,
	0.045742, 0.051272, 0.056597, 0.061708,
	0.066598, 0.071261, 0.075689, 0.079877,
	0.083819, 0.087511, 0.090949, 0.094129,
	0.097048, 0.099704, 0.1021, 0.10422,
	0.10608, 0.10767, 0.109, 0.11006,
	0.11086, 0.1114, 0.11168, 0.1117,
	0.11148, 0.111, 0.11029, 0.10933,
	0.10815, 0.10674, 0.10511, 0.10326,
	0.10121, 0.098965, 0.096527, 0.093906,
	0.091111, 0.088151, 0.085034, 0.081769,
	0.078366, 0.074834, 0.071182, 0.067421,
	0.063559, 0.059608, 0.055577, 0.051475,
	0.047313, 0.043101, 0.038849, 0.034567,
	0.030264, 0.025951, 0.021637, 0.017331,
	0.013043, 0.008783, 0.0045589, 0.00037992,
	-0.0037453, -0.0078082, -0.011801, -0.015715,
	-0.019543, -0.023278, -0.026913, -0.030441,
	-0.033855, -0.037149, -0.040318, -0.043356,
	-0.046258, -0.04902, -0.051636, -0.054104,
	-0.056419, -0.058578, -0.060579, -0.062418,
	-0.064095, -0.065607, -0.066952, -0.068131,
	-0.069142, -0.069986, -0.070662, -0.071171,
	-0.071515, -0.071693, -0.071709, -0.071563,
	-0.071259, -0.070799, -0.070186, -0.069423,
	-0.068513, -0.067461, -0.06627, -0.064946,
	-0.063491, -0.061913, -0.060214, -0.058401,
	-0.056479, -0.054453, -0.05233, -0.050115,
	-0.047814, -0.045434, -0.04298, -0.04046,
	-0.037879, -0.035243, -0.03256, -0.029837,
	-0.027078, -0.024292, -0.021485, -0.018663,
	-0.015832, -0.013, -0.010172, -0.0073545,
	-0.0045541, -0.0017767, 0.00097182, 0.0036857,
	0.0063592, 0.0089869, 0.011563, 0.014084,
	0.016543, 0.018935, 0.021258, 0.023505,
	0.025673, 0.027758, 0.029757, 0.031665,
	0.03348, 0.035199, 0.036818, 0.038336,
	0.039751, 0.041059, 0.042261, 0.043353,
	0.044336, 0.045208, 0.045968, 0.046616,
	0.047153, 0.047577, 0.047889, 0.04809,
	0.048182, 0.048163, 0.048037, 0.047805,
	0.047468, 0.047028, 0.046488, 0.04585,
	0.045116, 0.044289, 0.043372, 0.042369,
	0.041281, 0.040114, 0.03887, 0.037553,
	0.036167, 0.034716, 0.033203, 0.031633,
	0.03001, 0.028338, 0.026622, 0.024866,
	0.023074, 0.02125, 0.0194, 0.017527,
	0.015636, 0.013732, 0.011818, 0.0098996,
	0.0079806, 0.0060655, 0.0041584, 0.0022636,
	0.00038505, -0.0014732, -0.0033072, -0.0051131,
	-0.0068872, -0.008626, -0.010326, -0.011984,
	-0.013596, -0.015159, -0.016672, -0.01813,
	-0.019531, -0.020872, -0.022152, -0.023368,
	-0.024518, -0.025601, -0.026614, -0.027556,
	-0.028427, -0.029224, -0.029947, -0.030595,
	-0.031168, -0.031665, -0.032086, -0.03243,
	-0.032699, -0.032892, -0.03301, -0.033053,
	-0.033023, -0.032919, -0.032744, -0.032498,
	-0.032183, -0.0318, -0.031352, -0.030839,
	-0.030263, -0.029627, -0.028933, -0.028183,
	-0.027379, -0.026524, -0.02562, -0.02467,
	-0.023677, -0.022642, -0.02157, -0.020463,
	-0.019323, -0.018154, -0.016959, -0.015741,
	-0.014502, -0.013246, -0.011975, -0.010694,
	-0.0094039, -0.0081088, -0.0068114, -0.0055146,
	-0.0042215, -0.0029347, -0.0016571, -0.00039149,
	0.00085947, 0.0020932, 0.0033071, 0.0044987,
	0.0056656, 0.0068055, 0.0079163, 0.0089957,
	0.010042, 0.011052, 0.012026, 0.012961,
	0.013855, 0.014707, 0.015516, 0.016281,
	0.016999, 0.01767, 0.018294, 0.018869,
	0.019395, 0.01987, 0.020295, 0.02067,
	0.020993, 0.021265, 0.021486, 0.021657,
	0.021776, 0.021845, 0.021864, 0.021833,
	0.021754, 0.021626, 0.021452, 0.021231,
	0.020966, 0.020656, 0.020304, 0.01991,
	0.019476, 0.019004, 0.018495, 0.01795,
	0.017372, 0.016761, 0.01612, 0.015451,
	0.014755, 0.014034, 0.01329, 0.012526,
	0.011742, 0.010941, 0.010125, 0.0092957,
	0.0084555, 0.0076062, 0.0067497, 0.0058881,
	0.0050234, 0.0041574, 0.0032921, 0.0024295,
	0.0015714, 0.00071962, -0.00012395, -0.00095756,
	-0.0017795, -0.002588, -0.0033816, -0.0041585,
	-0.0049174, -0.0056567, -0.006375, -0.0070711,
	-0.0077436, -0.0083913, -0.0090132, -0.0096082,
	-0.010175, -0.010714, -0.011222, -0.011701,
	-0.012148, -0.012564, -0.012948, -0.0133,
	-0.013619, -0.013904, -0.014157, -0.014376,
	-0.014561, -0.014714, -0.014833, -0.014919,
	-0.014972, -0.014993, -0.014981, -0.014938,
	-0.014864, -0.014759, -0.014624, -0.01446,
	-0.014268, -0.014048, -0.013801, -0.013528,
	-0.01323, -0.012909, -0.012564, -0.012198,
	-0.01181, -0.011404, -0.010979, -0.010536,
	-0.010078, -0.0096053, -0.0091191, -0.0086207,
	-0.0081116, -0.0075928, -0.0070659, -0.006532,
	-0.0059925, -0.0054486, -0.0049017, -0.0043529,
	-0.0038035, -0.0032548, -0.002708, -0.0021642,
	-0.0016245, -0.0010903, -0.00056238, -4.2025e-005,
	0.0004698, 0.00097207, 0.0014638, 0.0019441,
	0.002412, 0.0028667, 0.0033074, 0.0037332,
	0.0041434, 0.0045374, 0.0049144, 0.0052738,
	0.0056152, 0.0059379, 0.0062415, 0.0065255,
	0.0067897, 0.0070336, 0.007257, 0.0074597,
	0.0076415, 0.0078022, 0.0079419, 0.0080605,
	0.008158, 0.0082345, 0.0082901, 0.008325,
	0.0083395, 0.0083337, 0.0083081, 0.0082628,
	0.0081984, 0.0081152, 0.0080136, 0.0078942,
	0.0077575, 0.007604, 0.0074342, 0.0072487,
	0.0070483, 0.0068334, 0.0066047, 0.006363,
	0.0061088, 0.005843, 0.0055661, 0.005279,
	0.0049823, 0.0046769, 0.0043634, 0.0040427,
	0.0037155, 0.0033826, 0.0030449, 0.0027029,
	0.0023577, 0.0020099, 0.0016604, 0.0013099,
	0.00095913, 0.00060893, 0.00026005, -8.6808e-005,
	-0.00043093, -0.00077158, -0.0011081, -0.0014398,
	-0.0017661, -0.0020863, -0.0023999, -0.0027064,
	-0.0030051, -0.0032955, -0.0035773, -0.0038499,
	-0.0041129, -0.004366, -0.0046088, -0.004841,
	-0.0050622, -0.0052722, -0.0054708, -0.0056577,
	-0.0058327, -0.0059957, -0.0061465, -0.006285,
	-0.0064111, -0.0065247, -0.0066258, -0.0067144,
	-0.0067904, -0.0068539, -0.0069049, -0.0069435,
	-0.0069698, -0.0069839, -0.006986, -0.0069763,
	-0.0069549, -0.0069221, -0.0068781, -0.0068232,
	-0.0067576, -0.0066818, -0.006596, -0.0065006,
	-0.006396, -0.0062825, -0.0061605, -0.0060306,
	-0.005893, -0.0057482, -0.0055967, -0.0054388,
	-0.0052751, -0.0051058, -0.0049315, -0.0047527,
	-0.0045696, -0.0043828, -0.0041926, -0.0039995,
	-0.0038038, -0.0036058, -0.0034062, -0.0032051,
	-0.0030029, -0.0028001, -0.0025969, -0.0023938,
	-0.0021911, -0.0019891, -0.0017882, -0.0015887,
	-0.001391, -0.0011953, -0.0010021, -0.00081155,
	-0.00062408, -0.00043984, -0.00025933, -8.2619e-005,
	8.9978e-005, 0.00025803, 0.00042169, 0.00058038,
	0.00073402, 0.00088249, 0.0010257, 0.0011633,
	0.0012953, 0.0014217, 0.0015423, 0.0016571,
	0.001766, 0.0018691, 0.0019663, 0.0020578,
	0.0021434, 0.0022233, 0.0022975, 0.002366,
	0.0024291, 0.0024866, 0.0025387, 0.0025854,
	0.0026269, 0.0026631, 0.0026943, 0.0027203,
	0.0027414, 0.0027577, 0.0027687, 0.0027755,
	0.0027774, 0.0027746, 0.0027673, 0.0027558,
	0.0027399, 0.0027199, 0.0026957, 0.0026679,
	0.0026363, 0.0026013, 0.0025628, 0.0025212,
	0.0024766, 0.0024295, 0.0023799, 0.0023281,
	0.0022743, 0.0022187, 0.0021616, 0.0021034,
	0.002044, 0.0019841, 0.0019233, 0.0018625,
	0.0018011, 0.0017398, 0.0016795, 0.0016174,
	0.0015577, 0.001498, 0.0014384, 0.0013791,
	0.0013208, 0.0012634, 0.0012067, 0.0011503,
	0.0010947, 0.0010396, 0.00098539, 0.00093197,
	0.00087945, 0.00082761, 0.00077656, 0.0007262,
	0.00067673, 0.00062815, 0.00058068, 0.00053431,
	0.00048927, 0.00044552, 0.00040326, 0.0003625,
	0.00032345, 0.00028614, 0.00025079, 0.00084549
};



// Specs:
// FIR LPF
// 1 + 767 coefficients (the first one is an additionnal 0)
// Fc = 0.5*Fs/64
// Passband frequency range: 0 - 9*Fc/10
// Passband level: +36.12 dB
// Passband ripple: 0.08 dB
// Stopband frequency range: 31*Fc/20 - Fs/2
// Stopband level: -oo dB
// Stopband attenuation: -85 dB
const double	InterpPack::_fir_2x_coef_arr [InterpRate2x::IMPULSE_LEN] =
{
	0, 0.001742279734, 0.0003227324025, 0.0003504220639,
	0.0003784201858, 0.0004060523799, 0.000433560667, 0.000460242331,
	0.0004863371066, 0.0005111187936, 0.0005348296736, 0.0005567178399,
	0.000577003141, 0.000594868077, 0.0006105017978, 0.0006229741377,
	0.0006324574426, 0.000637955229, 0.0006397031985, 0.0006367227753,
	0.0006294231227, 0.0006168285476, 0.0005994721071, 0.000576164941,
	0.0005473310058, 0.0005113181524, 0.0004686499247, 0.0004176613044,
	0.000360179431, 0.000294828955, 0.0002239625091, 0.0001397615714,
	4.917075029e-005, -4.906454506e-005, -0.0001586730304, -0.0002759525972,
	-0.0004037611106, -0.0005403286683, -0.0006872050613, -0.0008432719479,
	-0.001009559165, -0.001185183727, -0.00137082501, -0.00156565289,
	-0.001770105534, -0.001983373564, -0.00220574127, -0.002436438721,
	-0.002675609502, -0.002922412142, -0.003176764646, -0.003437642969,
	-0.003704785626, -0.003977186249, -0.004254645554, -0.004536168589,
	-0.004821306638, -0.0051085353, -0.005397156872, -0.005686020159,
	-0.005975209561, -0.006262452042, -0.006546344761, -0.006827480141,
	-0.007102792819, -0.007372232611, -0.007633943637, -0.00788688517,
	-0.008129587034, -0.008360797821, -0.008579124816, -0.008783232512,
	-0.008971721235, -0.00914319924, -0.009296252118, -0.009429483812,
	-0.009541564083, -0.00963116255, -0.009697032537, -0.00973786063,
	-0.009752387548, -0.009739324335, -0.009697580358, -0.009626088825,
	-0.009523960257, -0.009390153133, -0.009223736719, -0.009023832733,
	-0.008790026698, -0.008521722857, -0.008218257461, -0.007878933678,
	-0.007504467131, -0.007093708517, -0.006647210362, -0.006165014615,
	-0.005647212846, -0.005094386491, -0.004506902567, -0.003885599968,
	-0.003231214068, -0.002544844667, -0.00182755147, -0.001080682372,
	-0.0003055993791, 0.0004960207406, 0.001322467315, 0.002171720624,
	0.003041773856, 0.003930379128, 0.004835267947, 0.005753857674,
	0.006683530599, 0.007621410207, 0.008564704475, 0.009510380936,
	0.01045533995, 0.01139613927, 0.0123295518, 0.01325220433,
	0.01416068595, 0.01505101153, 0.01591999786, 0.01676380471,
	0.01757851783, 0.01836076832, 0.01910650003, 0.01981223275,
	0.02047409904, 0.02108852463, 0.02165179807, 0.02216040787,
	0.02261085054, 0.02299979734, 0.02332394892, 0.0235801863,
	0.0237654538, 0.02387693283, 0.02391195908, 0.02386807742,
	0.0237429762, 0.02353457915, 0.02324103906, 0.0228608214,
	0.02239262438, 0.02183534433, 0.02118814096, 0.02045058738,
	0.01962257961, 0.01870419866, 0.01769581587, 0.01659837872,
	0.01541303206, 0.01414098157, 0.01278441212, 0.0113452081,
	0.009826081402, 0.008229855716, 0.006559784686, 0.004819464346,
	0.003012819169, 0.001144170699, -0.0007818426856, -0.002760242962,
	-0.004785736303, -0.006852724797, -0.008955278991, -0.01108715657,
	-0.01324187199, -0.01541268359, -0.01759261257, -0.01977441632,
	-0.02195062519, -0.02411358996, -0.02625557221, -0.02836865061,
	-0.03044473373, -0.03247564125, -0.03445323438, -0.036369256,
	-0.03821536833, -0.03998331928, -0.04166508141, -0.04325230996,
	-0.04473714713, -0.04611169335, -0.04736816632, -0.04849918852,
	-0.04949738927, -0.0503558772, -0.05106786765, -0.051627048,
	-0.05202737026, -0.05226326572, -0.05232951832, -0.0522214062,
	-0.05193460354, -0.05146539981, -0.05081054857, -0.04996742686,
	-0.04893390401, -0.04770850047, -0.04629031953, -0.04467920731,
	-0.04287556956, -0.04088051322, -0.03869581877, -0.03632406645,
	-0.03376841422, -0.03103276272, -0.02812175712, -0.02504086042,
	-0.02179601806, -0.01839411765, -0.01484270347, -0.01114988159,
	-0.007324699964, -0.003376659821, 0.000683918981, 0.004846165968,
	0.009098558205, 0.01342894118, 0.01782462363, 0.02227232605,
	0.0267583219, 0.03126837881, 0.0357878146, 0.04030152292,
	0.04479407263, 0.04924969032, 0.05365237905, 0.05798585405,
	0.06223368171, 0.06637930212, 0.07040613857, 0.0742975011,
	0.07803678014, 0.08160747482, 0.08499328207, 0.08817799363,
	0.09114574369, 0.09388102591, 0.09636868824, 0.09859394674,
	0.1005427553, 0.1022013392, 0.1035567907, 0.1045967585,
	0.1053096883, 0.1056848338, 0.1057122442, 0.1053829182,
	0.1046887854, 0.1036228114, 0.1021790258, 0.1003525021,
	0.09813949524, 0.09553742354, 0.09254493804, 0.08916193185,
	0.08538957045, 0.08123030902, 0.07668798651, 0.07176776518,
	0.06647614661, 0.06082100909, 0.05481167793, 0.04845881893,
	0.04177446192, 0.03477204034, 0.02746641332, 0.01987371878,
	0.0120114516, 0.003898510203, -0.004445050075, -0.01299785294,
	-0.02173724057, -0.03063949823, -0.03967960565, -0.04883163923,
	-0.05806849217, -0.06736218009, -0.07668371714, -0.08600331368,
	-0.09529038984, -0.104513663, -0.1136411693, -0.1226404465,
	-0.1314784946, -0.1401219954, -0.1485372711, -0.1566904448,
	-0.1645474937, -0.172074439, -0.1792372909, -0.1860022387,
	-0.1923357307, -0.1982046325, -0.2035761834, -0.2084182126,
	-0.2126992079, -0.2163884272, -0.2194558911, -0.221872701,
	-0.223610932, -0.2246437585, -0.2249456736, -0.2244923915,
	-0.2232611184, -0.2212305232, -0.2183808121, -0.214693907,
	-0.2101533743, -0.2047446886, -0.198455083, -0.1912737755,
	-0.1831919406, -0.1742028128, -0.1643016924, -0.1534860427,
	-0.1417554206, -0.1291116623, -0.1155587919, -0.1011030852,
	-0.085753021, -0.06951943703, -0.05241538944, -0.03445621409,
	-0.01565946649, 0.003954984002, 0.02436513506, 0.04554676387,
	0.06747348062, 0.09011675956, 0.1134461258, 0.1374289443,
	0.162030763, 0.1872151458, 0.212943884, 0.2391770576,
	0.2658730424, 0.2929886858, 0.3204793157, 0.3482989051,
	0.3764001754, 0.4047346019, 0.4332526757, 0.4619038651,
	0.4906368532, 0.5193995898, 0.5481394221, 0.5768031931,
	0.6053374807, 0.6336885973, 0.6618027691, 0.689626244,
	0.7171055053, 0.7441872751, 0.7708187431, 0.7969476598,
	0.8225225065, 0.847492502, 0.8718078895, 0.8954199882,
	0.9182812855, 0.9403455873, 0.9615681762, 0.981905798,
	1.001316998, 1.019761942, 1.037202776, 1.053603525,
	1.068930331, 1.08315147, 1.096237409, 1.108160957,
	1.118897291, 1.128423998, 1.136721226, 1.143771574,
	1.149560292, 1.154075265, 1.157307047, 1.159248818,
	1.159896529, 1.159248818, 1.157307047, 1.154075265,
	1.149560292, 1.143771574, 1.136721226, 1.128423998,
	1.118897291, 1.108160957, 1.096237409, 1.08315147,
	1.068930331, 1.053603525, 1.037202776, 1.019761942,
	1.001316998, 0.981905798, 0.9615681762, 0.9403455873,
	0.9182812855, 0.8954199882, 0.8718078895, 0.847492502,
	0.8225225065, 0.7969476598, 0.7708187431, 0.7441872751,
	0.7171055053, 0.689626244, 0.6618027691, 0.6336885973,
	0.6053374807, 0.5768031931, 0.5481394221, 0.5193995898,
	0.4906368532, 0.4619038651, 0.4332526757, 0.4047346019,
	0.3764001754, 0.3482989051, 0.3204793157, 0.2929886858,
	0.2658730424, 0.2391770576, 0.212943884, 0.1872151458,
	0.162030763, 0.1374289443, 0.1134461258, 0.09011675956,
	0.06747348062, 0.04554676387, 0.02436513506, 0.003954984002,
	-0.01565946649, -0.03445621409, -0.05241538944, -0.06951943703,
	-0.085753021, -0.1011030852, -0.1155587919, -0.1291116623,
	-0.1417554206, -0.1534860427, -0.1643016924, -0.1742028128,
	-0.1831919406, -0.1912737755, -0.198455083, -0.2047446886,
	-0.2101533743, -0.214693907, -0.2183808121, -0.2212305232,
	-0.2232611184, -0.2244923915, -0.2249456736, -0.2246437585,
	-0.223610932, -0.221872701, -0.2194558911, -0.2163884272,
	-0.2126992079, -0.2084182126, -0.2035761834, -0.1982046325,
	-0.1923357307, -0.1860022387, -0.1792372909, -0.172074439,
	-0.1645474937, -0.1566904448, -0.1485372711, -0.1401219954,
	-0.1314784946, -0.1226404465, -0.1136411693, -0.104513663,
	-0.09529038984, -0.08600331368, -0.07668371714, -0.06736218009,
	-0.05806849217, -0.04883163923, -0.03967960565, -0.03063949823,
	-0.02173724057, -0.01299785294, -0.004445050075, 0.003898510203,
	0.0120114516, 0.01987371878, 0.02746641332, 0.03477204034,
	0.04177446192, 0.04845881893, 0.05481167793, 0.06082100909,
	0.06647614661, 0.07176776518, 0.07668798651, 0.08123030902,
	0.08538957045, 0.08916193185, 0.09254493804, 0.09553742354,
	0.09813949524, 0.1003525021, 0.1021790258, 0.1036228114,
	0.1046887854, 0.1053829182, 0.1057122442, 0.1056848338,
	0.1053096883, 0.1045967585, 0.1035567907, 0.1022013392,
	0.1005427553, 0.09859394674, 0.09636868824, 0.09388102591,
	0.09114574369, 0.08817799363, 0.08499328207, 0.08160747482,
	0.07803678014, 0.0742975011, 0.07040613857, 0.06637930212,
	0.06223368171, 0.05798585405, 0.05365237905, 0.04924969032,
	0.04479407263, 0.04030152292, 0.0357878146, 0.03126837881,
	0.0267583219, 0.02227232605, 0.01782462363, 0.01342894118,
	0.009098558205, 0.004846165968, 0.000683918981, -0.003376659821,
	-0.007324699964, -0.01114988159, -0.01484270347, -0.01839411765,
	-0.02179601806, -0.02504086042, -0.02812175712, -0.03103276272,
	-0.03376841422, -0.03632406645, -0.03869581877, -0.04088051322,
	-0.04287556956, -0.04467920731, -0.04629031953, -0.04770850047,
	-0.04893390401, -0.04996742686, -0.05081054857, -0.05146539981,
	-0.05193460354, -0.0522214062, -0.05232951832, -0.05226326572,
	-0.05202737026, -0.051627048, -0.05106786765, -0.0503558772,
	-0.04949738927, -0.04849918852, -0.04736816632, -0.04611169335,
	-0.04473714713, -0.04325230996, -0.04166508141, -0.03998331928,
	-0.03821536833, -0.036369256, -0.03445323438, -0.03247564125,
	-0.03044473373, -0.02836865061, -0.02625557221, -0.02411358996,
	-0.02195062519, -0.01977441632, -0.01759261257, -0.01541268359,
	-0.01324187199, -0.01108715657, -0.008955278991, -0.006852724797,
	-0.004785736303, -0.002760242962, -0.0007818426856, 0.001144170699,
	0.003012819169, 0.004819464346, 0.006559784686, 0.008229855716,
	0.009826081402, 0.0113452081, 0.01278441212, 0.01414098157,
	0.01541303206, 0.01659837872, 0.01769581587, 0.01870419866,
	0.01962257961, 0.02045058738, 0.02118814096, 0.02183534433,
	0.02239262438, 0.0228608214, 0.02324103906, 0.02353457915,
	0.0237429762, 0.02386807742, 0.02391195908, 0.02387693283,
	0.0237654538, 0.0235801863, 0.02332394892, 0.02299979734,
	0.02261085054, 0.02216040787, 0.02165179807, 0.02108852463,
	0.02047409904, 0.01981223275, 0.01910650003, 0.01836076832,
	0.01757851783, 0.01676380471, 0.01591999786, 0.01505101153,
	0.01416068595, 0.01325220433, 0.0123295518, 0.01139613927,
	0.01045533995, 0.009510380936, 0.008564704475, 0.007621410207,
	0.006683530599, 0.005753857674, 0.004835267947, 0.003930379128,
	0.003041773856, 0.002171720624, 0.001322467315, 0.0004960207406,
	-0.0003055993791, -0.001080682372, -0.00182755147, -0.002544844667,
	-0.003231214068, -0.003885599968, -0.004506902567, -0.005094386491,
	-0.005647212846, -0.006165014615, -0.006647210362, -0.007093708517,
	-0.007504467131, -0.007878933678, -0.008218257461, -0.008521722857,
	-0.008790026698, -0.009023832733, -0.009223736719, -0.009390153133,
	-0.009523960257, -0.009626088825, -0.009697580358, -0.009739324335,
	-0.009752387548, -0.00973786063, -0.009697032537, -0.00963116255,
	-0.009541564083, -0.009429483812, -0.009296252118, -0.00914319924,
	-0.008971721235, -0.008783232512, -0.008579124816, -0.008360797821,
	-0.008129587034, -0.00788688517, -0.007633943637, -0.007372232611,
	-0.007102792819, -0.006827480141, -0.006546344761, -0.006262452042,
	-0.005975209561, -0.005686020159, -0.005397156872, -0.0051085353,
	-0.004821306638, -0.004536168589, -0.004254645554, -0.003977186249,
	-0.003704785626, -0.003437642969, -0.003176764646, -0.002922412142,
	-0.002675609502, -0.002436438721, -0.00220574127, -0.001983373564,
	-0.001770105534, -0.00156565289, -0.00137082501, -0.001185183727,
	-0.001009559165, -0.0008432719479, -0.0006872050613, -0.0005403286683,
	-0.0004037611106, -0.0002759525972, -0.0001586730304, -4.906454506e-005,
	4.917075029e-005, 0.0001397615714, 0.0002239625091, 0.000294828955,
	0.000360179431, 0.0004176613044, 0.0004686499247, 0.0005113181524,
	0.0005473310058, 0.000576164941, 0.0005994721071, 0.0006168285476,
	0.0006294231227, 0.0006367227753, 0.0006397031985, 0.000637955229,
	0.0006324574426, 0.0006229741377, 0.0006105017978, 0.000594868077,
	0.000577003141, 0.0005567178399, 0.0005348296736, 0.0005111187936,
	0.0004863371066, 0.000460242331, 0.000433560667, 0.0004060523799,
	0.0003784201858, 0.0003504220639, 0.0003227324025, 0.001742279734
};



}	// namespace rspl



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
